"""
    Binary(θ)

Binary layer, with external fields `θ`.
"""
struct Binary{N, T, A <: AbstractArray{T,N}} <: AbstractLayer{N}
    θ::A
end
Binary(::Type{T}, n::Int...) where {T} = Binary(zeros(T, n...))
Binary(n::Int...) = Binary(Float64, n...)

Flux.@functor Binary

function effective(layer::Binary, inputs::AbstractTensor; β::Real = true)
    check_size(layer, inputs)
    return Binary(β * (layer.θ .+ inputs))
end

free_energies(layer::Binary) = -LogExpFunctions.log1pexp.(layer.θ)
transfer_mode(layer::Binary) = map(binary_mode, layer.θ)
transfer_mean(layer::Binary) = LogExpFunctions.logistic.(layer.θ)
transfer_mean_abs(layer::Binary) = transfer_mean(layer)
transfer_var(layer::Binary) = binary_var.(layer.θ)
transfer_std(layer::Binary) = binary_std.(layer.θ)

function transfer_sample(layer::Binary)
    u = rand(eltype(layer.θ), size(layer))
    return map(binary_rand, layer.θ, u)
end

function binary_mode(θ::Real)
    @assert !isnan(θ)
    return θ > 0
end

function binary_var(θ::Real)
    t = exp(-abs(θ))
    return t / (1 + t)^2
end

function binary_std(θ::Real)
    t = exp(-abs(θ))
    return √t / (1 + t)
end

function binary_rand(θ::Real, u::Real)
    @assert !isnan(θ) && !isnan(u)
    t = exp(-abs(θ))
    if θ ≥ 0
        return u * (1 + t) < 1
    else
        return u * (1 + t) < t
    end
end
